//
//	FLDLINK.CPP
//	
//	Functions to parse a string into a folder path, create the entry ID list
//	corresponding to a folder path, and write the entry ID list into a
//	folder link file (.XNK format) for Microsoft Exchange.
//
//	The following information on links to messages and folders is not part
//	of the MAPI specification. It applies only to the Microsoft Exchange
//	client version 4.0. Other MAPI clients, as well as future versions
//	of the Microsoft Exchange client, may not be compatible with links
//	built according to the information supplied here.
//

#include <windows.h>
#include <mapix.h>
#include <mapiutil.h>
#include <string.h>
#include "_links.h"

//
//	TraceFolderPath returns in *ppel a list of entry IDs, one for each node
//	in szPath. The first node names the message store; if it does not
//	match any entry in the message stores table, it assumes the path
//	refers to folders in the default message store.
//
//	This function destroys szPath by placing nulls in it.
//
//	The path is always assumed to begin at the store's IPM folder root.
//	It could of course be modified to include searching for folders
//	outside the store's IPM subtree; in that case, the IPM root would
//	have to be explicitly included in the path.
//
STDAPI
TraceFolderPath(LPMAPISESSION pses, LPSTR szPath, LPENTRYLIST FAR *ppel,
	LPMDB FAR *ppmdb)
{
	HRESULT			hr;
	LPMAPITABLE		ptbl = NULL;
	LPMDB			pmdb = NULL;
	LPMAPIFOLDER	pfld = NULL;
	ULONG			cbeid;
	LPENTRYID		peid;
	ULONG			ul;
	LPSTR			sz;
	int				cel;
	int				iel;
	LPSPropValue	pval = NULL;
	LPENTRYLIST		pel = NULL;

	*ppel = NULL;

	//	Count the nodes in szPath, add one for the store (which is optional),
	//	and allocate the output list of entry IDs.
	for (cel = 2, sz = szPath; sz; sz = strpbrk(sz, szPathSep))
	{
		++cel;
		++sz;
	}
	if ((hr = ResultFromScode(MAPIAllocateBuffer(sizeof(ENTRYLIST),
			(LPVOID FAR *) &pel)))
		|| (hr = ResultFromScode(MAPIAllocateMore(cel*sizeof(SBinary),
			(LPVOID) pel, (LPVOID FAR *) &pel->lpbin))))
		goto ret;
	ZeroMemory(pel->lpbin, cel*sizeof(SBinary));

	//	Look for first node in the list of message stores. If
	//	successful, this will place the store's entry ID in
	//	pel->lpbin[0].
	if (hr = pses->GetMsgStoresTable(0, &ptbl))
		goto ret;
	if (hr = TraceFolderNode(ptbl, &szPath, &pel->lpbin[0], pel))
		goto ret;
	UlRelease(ptbl);
	ptbl = NULL;

	//	Open the message store and find its IPM root folder.
	if (HR_FAILED(hr = pses->OpenMsgStore(0, pel->lpbin[0].cb, 
		(LPENTRYID) pel->lpbin[0].lpb, NULL, 
			MDB_NO_DIALOG | MAPI_DEFERRED_ERRORS, &pmdb)))
		goto ret;
	if (hr = HrGetOneProp((LPMAPIPROP) pmdb, PR_IPM_SUBTREE_ENTRYID, &pval))
		goto ret;
	cbeid = pval->Value.bin.cb;
	peid = (LPENTRYID) pval->Value.bin.lpb;
	if (hr = pmdb->OpenEntry(cbeid, peid, NULL, MAPI_DEFERRED_ERRORS, 
			&ul, (LPUNKNOWN FAR *) &pfld))
		goto ret;

	//	Loop through the remainder of the path, beginning with the
	//	IPM subtree (which is not explicitly named).
	for (iel = 1; szPath && *szPath; ++iel)
	{
		//	Open the folder and its hierarchy table.
		if (hr = pfld->GetHierarchyTable(MAPI_DEFERRED_ERRORS, &ptbl))
			goto ret;

		//	Search for the current node among the children of this folder.
		//	Bail out if it is not found.
		if (hr = TraceFolderNode(ptbl, &szPath, &pel->lpbin[iel], pel))
			goto ret;

		//	We just found the folder. Adjust variables so that we'll 
		//	open it on the next iteration.
		cbeid = pel->lpbin[iel].cb;
		peid = (LPENTRYID) pel->lpbin[iel].lpb;

		//	Clean up from this iteration...
		UlRelease(ptbl);
		ptbl = NULL;
		UlRelease(pfld);
		pfld = NULL;
		MAPIFreeBuffer(pval);
		pval = NULL;

		//	Convert short-term to long-term entry ID, overwriting the 
		//	element at iel. The memory is still linked and will still
		//	be cleaned up when pel is freed.
		if (hr = pmdb->OpenEntry(cbeid, peid, NULL, MAPI_DEFERRED_ERRORS, 
				&ul, (LPUNKNOWN FAR *) &pfld))
			goto ret;
		if (hr = HrGetOneProp((LPMAPIPROP) pfld, PR_ENTRYID, &pval))
			goto ret;
		if (hr = DupBin(&pel->lpbin[iel], &pval->Value.bin, pel))
			goto ret;
	}

	//	Return the list of entry IDs.
	pel->cValues = iel;
	*ppel = pel;
	if (ppmdb)
	{
		//	Return our pmdb and prevent its release.
		*ppmdb = pmdb;
		pmdb = NULL;
	}

ret:
	UlRelease(ptbl);
	UlRelease(pfld);
	UlRelease(pmdb);
	MAPIFreeBuffer(pval);
	if (hr)
		MAPIFreeBuffer(pel);
	
	return hr;
}

//
//	TraceFolderNode finds the first node of *pszPath in the table
//	specified in ptbl. It handles two subtly different cases: ptbl is
//	a folder hierarchy table, or ptbl is the message stores table.
//
//	The folder or store's entry ID is copied to pbin. The memory allocated
//	to hold the entry ID is chained to pv, so there is no need to free it
//	separately.
//
//	If the first node of *pszPath is used, *pszPath is advanced to point
//	to the next node and a null is left; otherwise it is unchanged.
//
STDAPI	
TraceFolderNode(LPMAPITABLE ptbl, LPSTR FAR *pszPath, LPSBinary pbin,
	LPVOID pv)
{
	HRESULT			hr;
	LPSRowSet		prows = NULL;
	int				irow;
	int				irowDefault = -1;
	int				irowToUse;
	LPSTR			sz;
	LPSTR			szNode;
	char			ch = 0;
	static SizedSPropTagArray(3, cols) = 
					{ 3, { PR_DISPLAY_NAME, PR_ENTRYID, PR_RESOURCE_FLAGS } };
	enum { iName = 0, iEntryID, iFlags };

	//	Snip the first node out of the path. Save the character where we
	//	stick the null, because we may need to restore it before returning.
	szNode = *pszPath;
	sz = strpbrk(szNode, szPathSep);
	if (sz)
	{
		ch = *sz;
		*sz++ = 0;
	}

	//	Retrieve the entire table. It may be lengthy for some message stores,
	//	such as an MS Exchange public folder store.
	if (hr = HrQueryAllRows(ptbl, (LPSPropTagArray) &cols, NULL, NULL,
			0, &prows))
		goto ret;

	for (irow = 0; irow < (int) prows->cRows; ++irow)
	{
		//	Slightly subtle: this column appears only in the 
		//	message stores table. If we have a folder hierarchy table,
		//	PR_RESOURCE_FLAGS will always be absent, and this functio
		//	will fail unless we find the named folder.
		if (prows->aRow[irow].lpProps[iFlags].ulPropTag == PR_RESOURCE_FLAGS &&
				(prows->aRow[irow].lpProps[iFlags].Value.ul & STATUS_DEFAULT_STORE))
			irowDefault = irow;

		//	Check for the folder we are looking for.
		//	Note the case insensitive comparison on the folder name.
		if (!lstrcmpi(szNode, prows->aRow[irow].lpProps[iName].Value.lpszA))
			break;
	}

	//	If we did not find what we were looking for, fail now.
	irowToUse = irow < (int) prows->cRows ? irow : irowDefault;
	if (irowToUse == -1)
	{
		hr = ResultFromScode(MAPI_E_NOT_FOUND);
		goto ret;
	}

	//	Copy the entry ID into the output list.
	if (hr = DupBin(pbin,
		&prows->aRow[irowToUse].lpProps[iEntryID].Value.bin,
			pv))
		goto ret;

	//	Adjust the szPath parameter to point to the next node.
	if (irow < (int) prows->cRows)
	{
		//	We used the node; move to the next one.
		*pszPath = sz;
	}
	else
	{
		//	We didn't use the first node; must have been the default store.
		//	Put the path string back the way it was.
		if (ch)
			*--sz = ch;
	}

ret:
	//	Clean up.
	FreeProws(prows);

	return hr;
}

//
//	DupBin copies one SBinary structure to another. The memory allocated
//	to hold the copied data is linked to the pv parameter.
//
STDAPI
DupBin(LPSBinary pDst, LPSBinary pSrc, LPVOID pv)
{
	HRESULT	hr;

	if (hr = ResultFromScode(MAPIAllocateMore(pSrc->cb, pv,
			(LPVOID FAR *)&pDst->lpb)))
		return hr;

	pDst->cb = pSrc->cb;
	CopyMemory(pDst->lpb, pSrc->lpb, pSrc->cb);
	return hrSuccess;
} 

//
//	WriteFolderLink formats folder link information in a buffer and places
//	it on the given stream using a single Write() call. The pel parameter
//	gives the path to the folder.
//
//	The stream is usually connected either to a .XNK file, or to the
//	PR_ATTACH_DATA_BIN property of a file attachment in a MAPI message.
//
STDAPI
WriteFolderLink(LPENTRYLIST pel, LPSTREAM pstream)
{
	HRESULT			hr;
	FLINK *			pflink;
	ULONG			cb;
	ULONG			cbW;
	int				cbPath;
	LPBYTE			pb = NULL;

	//	Calculate the size of the buffer and allocate it dynamically.
	cbPath = CbFromPel(pel);
	cb = sizeof(FLINK) + sizeof(ULONG) + cbPath;
	if (hr = ResultFromScode(MAPIAllocateBuffer((ULONG) cb,
			(LPVOID FAR *) &pb)))
		goto ret;
	pflink = (FLINK *)pb;

	//	Fill in header fields identifying this as a folder link file.
	pflink->ulVersion = FLINK_VERSION; 
	pflink->ulObjectType = MAPI_FOLDER; 
	pflink->cbIconData = 0; 

	//	Fill in fields that describe how the Exchange viewer should
	//	appear when displayed. This code calls for a window that takes
	//	up about 60% of a 640x480 display, and does not display the
	//	folder hierarchy or status bar. 
	//	Refer to the FVWIN structure in _links.h for full explanation
	//	of these fields and all possible values.
	pflink->fvwin.ulVersion = FVWIN_VERSION;
	pflink->fvwin.sw = SW_SHOWNORMAL;
	pflink->fvwin.xLeft = 64;
	pflink->fvwin.yTop = 48;
	pflink->fvwin.cx = 512;
	pflink->fvwin.cy = 384;

	pflink->fvwin.xSplit = 200;

	pflink->fvwin.fTree = 0;
	pflink->fvwin.fToolbar = 0xffffffff;
	pflink->fvwin.fStatusbar = 0;

	//	Place the entry ID list in the buffer.
	FlattenFolderPath(pel, pb + sizeof(FLINK), sizeof(ULONG) + cbPath);

	//	Write out the buffer to the stream in one go.
	if (hr = pstream->Write(pb, cb, &cbW))
		goto ret;
	if (cbW != cb)
	{
		hr = ResultFromScode(MAPI_E_NOT_ENOUGH_DISK);
		goto ret;
	}

ret:
	MAPIFreeBuffer(pb);

	return hr;

}

//	Assorted macros for flattening out the list of entry IDs.
//	Using an indexed array to compute the pad size is simpler and quicker
//	than doing modulo arithmetic.
static int rgpad[8] = { 4, 7, 6, 5, 4, 7, 6, 5 };
#define PADLENGTH(_cb)	(rgpad[((_cb) & 7)])
#define PUTLONG(_l) 	{ *((long *)pb) = (long)(_l); pb += sizeof(long); }
#define PUTSHORT(_s) 	{ *((short *)pb) = (short)(_s); pb += sizeof(short); }
#define PUTBLK(_p,_c)	{ memcpy(pb, (_p), (size_t)(_c)); pb += (_c); }

//	
//	CbFromPel computes the total size of the flattened entry ID list.
//
int CbFromPel(LPENTRYLIST pel)
{
	int			cb = 2;					// for the zero terminator
	int			cel;
	LPSBinary	pbin;

	for (cel = (int) pel->cValues, pbin = pel->lpbin; cel; --cel, ++pbin)
	{
		cb += 12						// 3 longs
			+ pbin->cb					// entry ID size
			+ PADLENGTH(pbin->cb);		// entry ID pad
	}

	return cb;
}

//
//	FlattenFolderPath takes a list of entry IDs and flattens it out to
//	a buffer in a special format required for folder link files.
//	The required buffer size cb can be obtained by calling CbFromPel().
//
//	The flattened form is as follows:
//		length of the whole list (NOT including count)
//		for each entry ID in the list:
//			length of the node (INCLUDING count)
//			entry ID type (MAPI_STORE or MAPI_FOLDER)
//			actual size of the entry ID
//			entry ID data
//			pad bytes (garbage fill, size determined by PADLENGTH macro)
//		two zero bytes as a terminator
//
void FlattenFolderPath(LPENTRYLIST pel, LPBYTE pb, int cb)
{
	int				cel;
	LPSBinary		pbin;
	long			lType = MAPI_STORE;

	//	First, write out the length of the entire list.
	memset(pb, 0, cb);
	PUTLONG(cb - sizeof(ULONG));

	for (cel = (int) pel->cValues, pbin = pel->lpbin; cel; --cel, ++pbin)
	{
		//	Write out the length of this element, including pad
		PUTLONG(12 + pbin->cb + PADLENGTH(pbin->cb));

		//	Write out the entry ID type.
		//	The first one is the store entry ID, and the rest are folders.
		PUTLONG(lType);
		lType = MAPI_FOLDER;

		//	Write out the true entry ID length.
		PUTLONG(pbin->cb);

		//	Write out the entry ID, and skip over the padding.
		PUTBLK(pbin->lpb, pbin->cb);
		pb += PADLENGTH(pbin->cb);
	}

	//	Finally, write out a 0 terminator for the list.
	PUTSHORT(0);
}

