//
//	MSGLINK.CPP
//
//	Functions for creating a link to a message in another message, in a format
//	that Microsoft Exchange will understand.
//
//	The following information on links to messages and folders is not part
//	of the MAPI specification. It applies only to the Microsoft Exchange
//	client version 4.0. Other MAPI clients, as well as future versions
//	of the Microsoft Exchange client, may not be compatible with links
//	built according to the information supplied here.
//

#include <windows.h>
#include <mapix.h>
#include <mapiutil.h>
#include "_links.h"

//	BUG in MAPIOID.H, chokes on __segname
#ifdef _WIN32
#define __based(a)
#endif
#define USES_OID_OLE2_STORAGE
#define INITOID
#include <mapioid.h>

//
//	TraceMessage returns in *ppval the long-term entry ID of a message
//	in pfldLink. It find the first message in that folder whose
//	*normalized* subject begins with szSubject.
//	
//	Actually, TWO properties are returned in *ppval: the subject, and
//	the long-term entry ID. Both are needed for the message link.
//	Indexes in the property value array are defined in _LINKS.H.
//
STDAPI
TraceMessage(LPMAPIFOLDER pfldLink, LPSTR szSubject, LPSPropValue FAR *ppval)
{
	HRESULT			hr;
	LPMAPITABLE		ptbl = NULL;
	LPMESSAGE		pmsg = NULL;
	LPSRowSet		prows = NULL;
	static SizedSPropTagArray(2, ptCols) =
					{ 2, { PR_NORMALIZED_SUBJECT, PR_ENTRYID } };
	static SizedSPropTagArray(2, ptMsg) =
					{ 2, { PR_SUBJECT, PR_ENTRYID } };
	static SizedSSortOrderSet(1, sso) = 
					{ 1, 0, 0, { PR_NORMALIZED_SUBJECT, TABLE_SORT_ASCEND } };
	SRestriction	res;
	SPropValue		val;
	LPSPropValue	pvalEIDShort;
	LPSPropValue	pvalOut = NULL;
	ULONG			ul;

	*ppval = NULL;

	//	Retrieve the folder contents table.
	if (hr = pfldLink->GetContentsTable(MAPI_DEFERRED_ERRORS, &ptbl))
		goto ret;
	if (hr = ptbl->SetColumns((LPSPropTagArray) &ptCols, TBL_BATCH))
		goto ret;
	if (hr = ptbl->SortTable((LPSSortOrderSet) &sso, TBL_BATCH))
		goto ret;

	//	Construct a property restriction for our subject.
	//	There are a few things to note here; the thrust is that this
	//	is an inexact search method which is OK for demo and lots of
	//	interactive uses, but inadequate if you want (let's say) to
	//	find the message whose subject exactly matches a saved value.
	//	1. We're searching on PR_NORMALIZED_SUBJECT, which excludes
	//	prefixes such as RE: and FW:. Not all message stores make
	//	PR_SUBJECT available in the contents table.
	//	2. We're using RELOP_GE, which allows the caller of this
	//	function to pass in an incomplete subject.
	//	3. Very long subjects are truncated (at 255 bytes) in the
	//	contents table, so total precision in matching the subject
	//	is impossible anyway.
	res.rt = RES_PROPERTY;
	res.res.resProperty.relop = RELOP_GE;
	res.res.resProperty.ulPropTag = PR_NORMALIZED_SUBJECT;
	res.res.resProperty.lpProp = &val;
	val.ulPropTag = PR_NORMALIZED_SUBJECT;
	val.Value.lpszA = szSubject;

	//	Find the first message with a matching subject.
	//	If there is none, FindRow will return MAPI_E_NOT_FOUND.
	if (HR_FAILED(hr = ptbl->FindRow(&res, BOOKMARK_BEGINNING, 0)))
		goto ret;
	if (hr = ptbl->QueryRows(1, 0, &prows))
		goto ret;
	if (prows->cRows == 0 || prows->aRow[0].lpProps[1].ulPropTag != PR_ENTRYID)
	{
		hr = ResultFromScode(MAPI_E_NOT_FOUND);
		goto ret;
	}
	pvalEIDShort = &prows->aRow[0].lpProps[1];

	//	Retrieve the subject and long-term entry ID. The latter is
	//	necessary if the entry ID is to be stored permanently anywhere,
	//	such as a link in another message. Take care of some
	//	error conditions.
	if (hr = pfldLink->OpenEntry(pvalEIDShort->Value.bin.cb,
		(LPENTRYID) pvalEIDShort->Value.bin.lpb, NULL, MAPI_DEFERRED_ERRORS,
			&ul, (LPUNKNOWN FAR *) &pmsg))
		goto ret;
	if (hr = pmsg->GetProps((LPSPropTagArray) &ptMsg, 0, &ul, &pvalOut))
		goto ret;
	if (pvalOut[ivalEntryID_TM].ulPropTag != PR_ENTRYID)
	{
		//	This should not be happening, but nothing is impossible.
		hr = ResultFromScode(pvalOut[1].Value.err);
		goto ret;
	}
	if (pvalOut[ivalSubject_TM].ulPropTag != PR_SUBJECT)
	{
		LPSTR	szT;

		//	This can happen if the message subject is huge.
		//	We could stream out the property; instead, we'll use
		//	the truncated version from the table.
		szT = prows->aRow[0].lpProps[0].Value.lpszA;
		ul = prows->aRow[0].lpProps[0].ulPropTag == PR_NORMALIZED_SUBJECT ?
			lstrlenA(szT) + 1 : 0;
		if (ul == 0)
		{
			//	Last chance: hard-coded string.
			szT = "(no subject)";
			ul = lstrlenA(szT) + 1;
		}
		if (ul && !(hr = ResultFromScode(MAPIAllocateMore(ul, pvalOut,
			(LPVOID FAR *) &pvalOut[ivalSubject_TM].Value.lpszA))))
			goto ret;
		lstrcpyA(pvalOut[ivalSubject_TM].Value.lpszA, szT);
	}

	*ppval = pvalOut;

ret:
	//	Clean up after ourselves.
	UlRelease(pmsg);
	UlRelease(ptbl);
	FreeProws(prows);
	if (hr)
		MAPIFreeBuffer(pvalOut);

	return hr;
}

//
//	WriteMessageLink creates an attachment in pmsg which Microsoft
//	Exchange will recognize as a link to another message. That message's
//	entry ID is in pbin, and the full path to that message (message store
//	and parent folder or folders) is in pel.
//	If non-null, szSubject becomes the caption for the attachment, and
//	hicon becomes its icon.
//
//	An alternate implementation of this function might take an open message
//	interface for the message to link, and retrieve the message store
//	and folder entry IDs from that interface.
//

STDAPI
WriteMessageLink(LPMESSAGE pmsg, LPSBinary pbinStore, LPSBinary pbinFolder,
	LPSBinary pbinMessage, LPSTR szSubject, HANDLE hicon, ULONG ulPos)
{
	HRESULT			hr;
	LPATTACH		patch = NULL;
	LPSTORAGE		pstg = NULL;
	ULONG			ul;
	SPropValue		rgval[4];
	LPSPropProblemArray pprobs = NULL;

	//	Supply a default subject if none was passed in.
	if (!szSubject || !*szSubject)
		szSubject = "(no subject)";

	//	Create a new attachment in the message, and a structured
	//	storage for the attachment's content.
	if (hr = pmsg->CreateAttach(NULL, MAPI_DEFERRED_ERRORS, &ul, &patch))
		goto ret;
	if (hr = patch->OpenProperty(PR_ATTACH_DATA_OBJ, (LPIID) &IID_IStorage,
		STGM_READWRITE | STGM_SHARE_EXCLUSIVE, MAPI_CREATE | MAPI_MODIFY,
			(LPUNKNOWN *) &pstg))
		goto ret;

	//	Create a stream in the structured storage for each data item.
	if (hr = WriteMessageLinkStg(pstg, pbinStore, pbinFolder, pbinMessage,
			szSubject, hicon))
		goto ret;

	//	Save the structured storage.
	if (HR_FAILED(hr = pstg->Commit(STGC_DEFAULT)))
		goto ret;

	//	Write out other necessary attachment properties.
	rgval[0].ulPropTag = PR_ATTACH_METHOD;
	rgval[0].Value.l = ATTACH_OLE;
	rgval[1].ulPropTag = PR_ATTACH_TAG;
	rgval[1].Value.bin.cb = CB_OID_OLE2_STORAGE;
	rgval[1].Value.bin.lpb = (LPBYTE) &OID_OLE2_STORAGE;
	rgval[2].ulPropTag = PR_RENDERING_POSITION;
	rgval[2].Value.ul = ulPos;
	rgval[3].ulPropTag = PR_DISPLAY_NAME;
	rgval[3].Value.lpszA = szSubject;
	if (hr = patch->SetProps(4, rgval, &pprobs))
		goto ret;
	if (pprobs)
	{
		hr = ResultFromScode(pprobs->aProblem[0].scode);
		goto ret;
	}

	//	Save the attachment.
	if (hr = patch->SaveChanges(0))
		goto ret;

ret:
	//	Clean up after ourselves.
	UlRelease(patch);
	UlRelease(pstg);

	return hr;
}

//
//	WriteMsgLinkStg takes all the pieces of a message link and
//	puts them in a structured storage, in the format expected by
//	Microsoft Exchange. The three pbin parameters are the entry IDs of
//	the message store and folder containing the message, and of the
//	message itself. The subject and icon, which are both optional,
//	are used to create a rendering for the link.
//

static OLECHAR	szStreamSubject[]	= OLESTR("MailMsgAttSubject");
static OLECHAR	szStreamMessageEID[]= OLESTR("MailMsgAttMsg");
static OLECHAR	szStreamFolderEID[]	= OLESTR("MailMsgAttFld");
static OLECHAR	szStreamStoreEID[]	= OLESTR("MailMsgAttMdb");
static OLECHAR	szStreamIcon[]		= OLESTR("MailMsgAttIcon");
static CLSID	clsidMessageLink	= { 0x00020D09, 0, 0,
									    { 0xC0, 0, 0, 0, 0, 0, 0, 0x46 } };

STDAPI
WriteMessageLinkStg(LPSTORAGE pstg, LPSBinary pbinStore, LPSBinary pbinFolder,
	LPSBinary pbinMessage, LPSTR szSubject, HANDLE hicon)
{
	HRESULT			hr;

	//	Supply a default subject if none was passed in.
	if (!szSubject || !*szSubject)
		szSubject = "(no subject)";

	//	Create four streams in the structured storage, containing:
	//		...the message's subject
	if (hr = PutStgStream(pstg, szStreamSubject, lstrlenA(szSubject) + 1,
			(LPBYTE) szSubject))
		goto ret;
	//		...the message's entry ID
	if (hr = PutStgStream(pstg, szStreamMessageEID, pbinMessage->cb,
			pbinMessage->lpb))
		goto ret;
	//		...the parent folder's entry ID
	if (hr = PutStgStream(pstg, szStreamFolderEID, pbinFolder->cb,
			pbinFolder->lpb))
		goto ret;
	//		...the message store's entry ID
	if (hr = PutStgStream(pstg, szStreamStoreEID, pbinStore->cb,
			pbinStore->lpb))
		goto ret;

	//	Create a picture for display in the message body, and save it
	//	to the storage.
	if (hicon)
	{
		if (hr = SetStgRendering(hicon, szSubject, pstg, szStreamIcon))
			goto ret;
	}

	//	Mark the storage with the appropriate classID, so Exchange
	//	can launch it.
	if (hr = WriteClassStg(pstg, clsidMessageLink))
		goto ret;

ret:
	//	Nothing to clean up.

	return hr;
}

//
//	LinkMessageToMessage creates an attachment in pmsgOuter which
//	Microsoft Exchange interprets as a link to pmsgLink. All the
//	properties needed for the link itself are retrieved from pmsgLink.
//	The hicon and ulPos parameters are used to create a rendering for
//	the link in pmsgOuter.
//

STDAPI
LinkMessageToMessage(LPMESSAGE pmsgLink, LPMESSAGE pmsgOuter, 
	HANDLE hicon, ULONG ulPos)
{
	HRESULT			hr;
	LPATTACH		patch = NULL;
	LPSTORAGE		pstg = NULL;
	LPSPropValue	pval = NULL;
	LPSPropProblemArray	pprobs = NULL;
	SPropValue		rgval[4];
	static SizedSPropTagArray(4, tagaLink) = 
					{ 4, { PR_SUBJECT, PR_ENTRYID, PR_PARENT_ENTRYID,
						   PR_STORE_ENTRYID } };
	ULONG			ul;
	LPSTR			szSubject;

	//	Retrieve the entry IDs we need, and optional message subject.
	hr = pmsgLink->GetProps((LPSPropTagArray) &tagaLink, 0, &ul, &pval);
	if (HR_FAILED(hr))
		goto ret;
	if (pval[1].ulPropTag != PR_ENTRYID ||
		pval[2].ulPropTag != PR_PARENT_ENTRYID ||
		pval[3].ulPropTag != PR_STORE_ENTRYID)
	{
		hr = ResultFromScode(E_UNEXPECTED);
		goto ret;
	}
	if (pval[0].ulPropTag == PR_SUBJECT)
		szSubject = pval[0].Value.lpszA;
	else
		szSubject = "";

	//	Create a new attachment in the message, and a structured
	//	storage for the attachment's content.
	if (hr = pmsgOuter->CreateAttach(NULL, MAPI_DEFERRED_ERRORS, &ul, &patch))
		goto ret;
	if (hr = patch->OpenProperty(PR_ATTACH_DATA_OBJ, (LPIID) &IID_IStorage,
		STGM_READWRITE | STGM_SHARE_EXCLUSIVE, MAPI_CREATE | MAPI_MODIFY,
			(LPUNKNOWN *) &pstg))
		goto ret;

	//	Create a stream in the structured storage for each data item.
	if (hr = WriteMessageLinkStg(pstg, &pval[3].Value.bin, 
			&pval[2].Value.bin, &pval[1].Value.bin, szSubject, hicon))
		goto ret;

	//	Save the structured storage.
	if (HR_FAILED(hr = pstg->Commit(STGC_DEFAULT)))
		goto ret;
	
	//	Write out other necessary attachment properties.
	rgval[0].ulPropTag = PR_ATTACH_METHOD;
	rgval[0].Value.l = ATTACH_OLE;
	rgval[1].ulPropTag = PR_ATTACH_TAG;
	rgval[1].Value.bin.cb = CB_OID_OLE2_STORAGE;
	rgval[1].Value.bin.lpb = (LPBYTE) &OID_OLE2_STORAGE;
	rgval[2].ulPropTag = PR_RENDERING_POSITION;
	rgval[2].Value.ul = ulPos;
	rgval[3].ulPropTag = PR_DISPLAY_NAME;
	rgval[3].Value.lpszA = szSubject;
	if (hr = patch->SetProps(4, rgval, &pprobs))
		goto ret;
	if (pprobs)
	{
		hr = ResultFromScode(pprobs->aProblem[0].scode);
		goto ret;
	}

	//	Save the attachment.
	if (hr = patch->SaveChanges(0))
		goto ret;

ret:
	//	Clean up after ourselves.
	UlRelease(pstg);
	UlRelease(patch);
	MAPIFreeBuffer(pval);
	MAPIFreeBuffer(pprobs);

	return hr;
}
